from moteur_sans_variables.chainage import Chainage
from moteur_avec_variables.filtre import Filtre

class ChainageAvantAvecVariables_FC(Chainage):
    """ Un moteur d'inférence à chaînage avant avec variables. """

    def __init__(self, connaissances, methode=None):
        """
            :param methode: ``Filtre`` ou ``Unificateur``, détermine le type de\
            pattern match à appliquer. ``Filtre`` par défaut.
        """

        Chainage.__init__(self, connaissances)

        if methode is None:
            self.methode = Filtre()
        else:
            self.methode = methode

    def instancie_conclusion(self, regle, envs_et_fcs):
        """ Instancie la conclusion d'une règle pour tous les environnements.

            :param regle: la règle dont la conclusion doit être instanciée.
            :param list envs_et_fcs: une liste de paires d'environnements et de\
            facteurs de certitude servant à instancier la conclusion de la règle.
            :return: une liste de propositions correspondant aux différentes\
            instanciations de la conclusion.
        """
        nouveaux_faits = []

        for env, env_fc in envs_et_fcs:
            prop = self.methode.substitue(regle.conclusion, env)
            nouveau_fait = (prop, max(0.0, env_fc) * regle.fc)
            nouveaux_faits.append(nouveau_fait)

        return nouveaux_faits

    def chaine(self):
        """ Effectue le chaînage avant sur les faits et les règles contenus\
            dans la base de connaissances.
        """
        queue = [p for p in self.connaissances.faits.items()]
        self.reinitialise()
        
        while len(queue) > 0:
            fait, fait_fc = queue.pop(0)

            if (fait, fait_fc) not in self.solutions:
                self.trace.append((fait, fait_fc))
                self.solutions.append((fait, fait_fc))

                # Si le facteur de certitude du fait est supérieur à 0.0
                if fait_fc > 0.0:

                    # on vérifie si des règles sont déclenchées par le nouveau fait.
                    for regle in self.connaissances.regles:
                        envs = regle.depend_de(fait, self.methode)
            
                        for cond, env in envs.items():
                            # On remplace l'environnement par ceux qui satisfont toutes les conditions de la règle
                            # et pas seulement la première condition.
                            envs1 = regle.satisfaite_par(self.solutions, cond, env, fait_fc, self.methode)
            
                            # On ajoute à la queue la conclusion de la règle instanciée selon chaque environnement possible.                        
                            if len(envs1) > 0:
                                queue.extend(self.instancie_conclusion(regle, envs1))
                                self.trace.append(regle)

        return self.solutions